# -*- coding: utf-8 -*-
'''PyGame driver.'''

import pygame
import logging

from tuikit.driver import Driver
from tuikit.common import Coords, Size, MouseEvent


class CharCache:
    
    '''CharCache should implement character cache, as the name suggests.
    
    Currently it does not, characters ale rendered directly from TTF.
    
    '''
    
    def __init__(self):
        fontselect = 'dejavusansmono,liberationmono,freemono'
        self.font = pygame.font.SysFont(fontselect, 14)
        self.font_bold = pygame.font.SysFont(fontselect, 14, True)
        # get advance of some random char (all should be same in monospace font)
        advance = self.font.metrics('Q')[0][4]
        height = self.font.get_height()
        self.charsize = Size(advance, height)
        self.ascent = self.font.get_ascent()
        
        # choose self.render() implementation
        if hasattr(self.font, 'render_glyph'):
            self.render = self.render_glyph
        else:
            self.render = self.render_noglyph

    def render_glyph(self, screen, x, y, c, fgcolor, bgcolor, attr):
        '''Render using render_glyph and metrics.
        
        This is the correct way, but the output seems same as of render_noglyph
        and this implementation requires patching PyGame to work.
        
        This implements render() method. See render_noglyph for other implementation.
        
        '''
        if attr == 'bold':
            font = self.font_bold
        else:
            font = self.font
        
        # render character, get metrics
        surface = font.render_glyph(c, True, fgcolor, bgcolor)
        metrics = font.metrics(c)[0]
        minx, maxx, miny, maxy, advance = metrics
        height, ascent = self.charsize.h, self.ascent
        
        # clips origin and area of rendered character according to metrics
        startx, starty = 0, 0
        if minx < 0:
            startx = abs(minx)
            minx = 0
        if maxy > ascent:
            starty = maxy - ascent
            maxy -= starty
        if ascent - miny > height:
            miny = ascent - height
        if maxx > advance:
            maxx = advance
        
        # draw background
        dest = Coords(x * self.charsize.w, y * self.charsize.h)
        screen.fill(bgcolor, pygame.Rect(dest.x, dest.y, self.charsize.w, self.charsize.h))
        
        # draw character
        dest.x += minx
        dest.y += ascent - maxy
        area = pygame.Rect(startx, starty, maxx - minx, maxy - miny)
        screen.blit(surface, tuple(dest), area)
    
    def render_noglyph(self, screen, x, y, c, fgcolor, bgcolor, attr):
        '''Render character using normal text rendering.
        
        This implements render() method. See render_glyph for other implementation.
        
        '''
        if attr == 'bold':
            font = self.font_bold
        else:
            font = self.font
               
        # render character, get metrics
        surface = font.render(c, True, fgcolor, bgcolor)
        metrics = font.metrics(c)[0]
        minx = metrics[0]
        startx = 0
        if minx < 0:
            startx = abs(minx)
        
        # draw background
        dest = Coords(x * self.charsize.w, y * self.charsize.h)
        screen.fill(bgcolor, pygame.Rect(dest.x, dest.y, self.charsize.w, self.charsize.h))
        
        # draw character
        area = pygame.Rect(startx, 0, self.charsize.w, self.charsize.h)
        screen.blit(surface, tuple(dest), area)


class DriverPygame(Driver):
    
    '''PyGame driver class.'''
    
    keymap = {
        pygame.K_ESCAPE     : 'escape',
        pygame.K_TAB        : 'tab',
        pygame.K_RETURN     : 'enter',
        pygame.K_BACKSPACE  : 'backspace',
        pygame.K_F1         : 'f1',
        pygame.K_F2         : 'f2',
        pygame.K_F3         : 'f3',
        pygame.K_F4         : 'f4',
        pygame.K_F5         : 'f5',
        pygame.K_F6         : 'f6',
        pygame.K_F7         : 'f7',
        pygame.K_F8         : 'f8',
        pygame.K_F9         : 'f9',
        pygame.K_F10        : 'f10',
        pygame.K_F11        : 'f11',
        pygame.K_F12        : 'f12',
        pygame.K_INSERT     : 'insert',
        pygame.K_DELETE     : 'delete',
        pygame.K_HOME       : 'home',
        pygame.K_END        : 'end',
        pygame.K_PAGEUP     : 'pageup',
        pygame.K_PAGEDOWN   : 'pagedown',
        pygame.K_UP         : 'up',
        pygame.K_DOWN       : 'down',
        pygame.K_LEFT       : 'left',
        pygame.K_RIGHT      : 'right',
        pygame.K_PRINT      : 'print',
        pygame.K_SCROLLOCK  : 'scrollock',
        pygame.K_PAUSE      : 'pause',
        }
    
    colormap = {
        'black'         : (0,0,0),
        'blue'          : (23,23,178),
        'green'         : (23,178,23),
        'cyan'          : (23,178,178),
        'red'           : (178,23,23),
        'magenta'       : (178,23,178),
        'yellow'        : (178,103,23),
        'white'         : (178,178,178),
        'intenseblack'  : (104,104,104),
        'intenseblue'   : (84,84,255),
        'intensegreen'  : (84,255,84),
        'intensecyan'   : (84,255,255),
        'intensered'    : (255,84,84),
        'intensemagenta': (255,84,255),
        'intenseyellow' : (255,255,84),
        'intensewhite'  : (255,255,255),
        }
    
    def __init__(self):
        '''Initialize instance attributes'''
        Driver.__init__(self)
        self.log = logging.getLogger('tuikit')
        self.screen = None
        self.size.w, self.size.h = 120, 40  # screen size in characters
        self.charcache = None
        self.charsize = Size(16, 8)  # character size in pixels
        self.last_keypress = None
        self.last_key = None
        self.colors = {}     # maps names to curses attributes
        self.colorstack = [] # pushcolor/popcolor puts or gets attributes from this
        self.default_color = (self.colormap['white'], self.colormap['black'], 0)
        self.current_color = self.default_color

    def start(self, mainfunc):
        pygame.init()
        self.charcache = CharCache()
        self.charsize = self.charcache.charsize
        mode = self.size.w * self.charsize.w, self.size.h * self.charsize.h
        self.screen = pygame.display.set_mode(mode, pygame.RESIZABLE)
        mainfunc()


    ## input ##
    
    def getevents(self, timeout=None):
        '''Process input, return list of events.'''
        events = []
        for ev in pygame.event.get():
            # mouse
            if ev.type == pygame.MOUSEMOTION:
                evdata = MouseEvent(ev.pos[0] // self.charsize.w, ev.pos[1] // self.charsize.h)
                events.append(('mousemove', evdata))
            elif ev.type in (pygame.MOUSEBUTTONDOWN, pygame.MOUSEBUTTONUP):
                evdata = MouseEvent(ev.pos[0] // self.charsize.w, ev.pos[1] // self.charsize.h)
                evdata.button = ev.button
                evname = {pygame.MOUSEBUTTONDOWN: 'mousedown', pygame.MOUSEBUTTONUP: 'mouseup'}
                events.append((evname[ev.type], evdata))
            
            # keyboard
            elif ev.type == pygame.KEYDOWN:
                keypress = self.keypress_from_pygame_event(ev)
                if keypress:
                    events.append(keypress)
                    self.last_keypress = keypress
                    self.last_key = ev.key
                    pygame.time.set_timer(pygame.USEREVENT, 200)
            elif ev.type == pygame.USEREVENT: # repeat last key press
                events.append(self.last_keypress)
                pygame.time.set_timer(pygame.USEREVENT, 50)
            elif ev.type == pygame.KEYUP:
                if ev.key == self.last_key:
                    pygame.time.set_timer(pygame.USEREVENT, 0)
            
            # window
            elif ev.type == pygame.VIDEORESIZE:
                neww, newh = ev.w // self.charsize.w, ev.h // self.charsize.h
                if neww != self.size.w or newh != self.size.h:
                    self.size.w, self.size.h = neww, newh
                    mode = self.size.w * self.charsize.w, self.size.h * self.charsize.h
                    self.screen = pygame.display.set_mode(mode, pygame.RESIZABLE)
                    events.append(('resize',))
            elif ev.type == pygame.QUIT:
                events.append(('quit',))
            
            else:
                self.log.warning('Unknown PyGame event: %r', ev.type)
        return events

    def keypress_from_pygame_event(self, ev):
        if ev.key in self.keymap:
            keypress = ('keypress', self.keymap[ev.key], None)
        elif ev.unicode:
            keypress = ('keypress', None, ev.unicode)
        else:
            self.log.debug('Unknown key: key=%r unicode=%r' % (ev.key, ev.unicode))
            keypress = None
        return keypress

    ## drawing ##
    
    def erase(self):
        '''Clear screen.'''
        self.screen.fill(self.colormap['black'])
    
    def putch(self, x, y, c):
        if not self.clipstack.test(x, y):
            return
        fgcolor, bgcolor, attr = self.current_color
        self.charcache.render(self.screen, x, y, c, fgcolor, bgcolor, attr)

    def commit(self):
        '''Commit changes to the screen.'''
        pygame.display.flip()


    ## colors ##
    
    def _parsecolor(self, name, attr=None):
        name = name.lower().strip()
        if attr == 'bold':
            name = 'intense' + name
        return self.colormap[name]
    
    def _parseattrs(self, attrs):
        res = ''
        for a in attrs:
            a = a.lower().strip()
            if a == 'bold':
                res = a
        return res

    def setcolor(self, name, desc):
        '''Define color name.
        
        name - name of color (e.g. 'normal', 'active')
        desc - color description - foreground, background, attributes (e.g. 'black on white, bold')
        
        '''
        parts = desc.split(',')
        fg, bg = parts[0].split(' on ')
        attrs = parts[1:]
        attr = self._parseattrs(attrs)
        fg = self._parsecolor(fg, attr)
        bg = self._parsecolor(bg)
        self.colors[name] = (fg, bg, attr)

    def pushcolor(self, name):
        '''Add color on top of stack and use this color for following output.'''
        # add prefix if such color is available
        if len(self.colorprefix):
            prefixname = self.colorprefix[-1] + name
            if prefixname in self.colors:
                name = prefixname
        col = self.colors[name]
        self.current_color = col
        self.colorstack.append(col)
    
    def popcolor(self):
        '''Remove color from top of stack and use new top color for following output.'''
        self.colorstack.pop()
        if len(self.colorstack):
            col = self.colorstack[-1]
        else:
            col = self.default_color
        self.current_color = col


    ## cursor ##

    def showcursor(self, x, y):
        '''Set cursor to be shown at x, y coordinates.'''

    def hidecursor(self):
        '''Hide cursor.'''

